
/** @defgroup GSH GSH - the gfx shell */
/** @{ */

/**
 * @file gsh.h
 *
 * GSH is the gfx shell.  It can have various front ends, which can either
 * be graphical or terminal-based.  GSH itself just implements the core
 * shell functionality, not the user interface.
 */


#ifndef GSH_H
#define GSH_H

#include <stdio.h>
#include <string.h>
#include "unwindows.h"
#include <stdlib.h>
#include <errno.h>

#ifdef __cplusplus
extern "C" {
#endif

/** @name Constants */
/** @{ */
#define GSH_VERSION "0.7"

#define MAX_HISTORY_LINES     30    /**< how big a scrollback buffer can be */
#define DEFAULT_LINE_LENGTH   81    /**< how long an input line can be */
#define CMD_DELINEATE         ';'   /**< command separator */
#define MAX_ARGS              10    /**< maximum command argv length */
/** @} */

/** @name Floating-point CVARs */
/** @{ */
/**
 * Create a floating-point shell variable.
 *
 * @note In order to use the shell variable, you still have to
 * tell GSH about it at runtime;
 * @see gsh_REGISTER()
 *
 * @param VARNAME the name of the variable
 * @param VALUE   the initial value to give the variable
 * @param HELP    the help text associated with the variable
 */
#define gsh_DECLARE_FCVAR_INIT_IMPL(VARNAME,VALUE)          \
    float   CVAR_##VARNAME = VALUE;                         \
    static void CMD_##VARNAME(int argc, char **argv)        \
    {                                                       \
        if (argc == 2)                                      \
        {                                                   \
            CVAR_##VARNAME = (float)atof(argv[1]);          \
        }                                                   \
        else if (argc > 2)                                  \
        {                                                   \
            gsh_Printf(#VARNAME " [value]\n");              \
            return;                                         \
        }                                                   \
        gsh_Printf(#VARNAME " is %f\n", CVAR_##VARNAME);    \
    }

#define gsh_DECLARE_FCVAR_INIT_EX(VARNAME,VALUE,HELP)       \
    const char* HELP_##VARNAME = "(f) : " HELP;             \
    gsh_DECLARE_FCVAR_INIT_IMPL(VARNAME,VALUE)

#define gsh_DECLARE_FCVAR_INIT(VARNAME,VALUE)               \
    const char* HELP_##VARNAME = "(f)";                     \
    gsh_DECLARE_FCVAR_INIT_IMPL(VARNAME,VALUE)

/**
 * Create a floating-point shell variable.
 *
 * @note In order to use the shell variable, you still have to
 * tell GSH about it at runtime;
 * @see gsh_REGISTER()
 *
 * @param VARNAME the name of the variable
 * @param HELP    the help text associated with the variable
 * @note the initial value is hardcoded to 0.0f
 */

#define gsh_DECLARE_FCVAR_EX(VARNAME,HELP)                  \
    gsh_DECLARE_FCVAR_INIT_EX(VARNAME,0.0f,HELP)

#define gsh_DECLARE_FCVAR(VARNAME)                          \
    gsh_DECLARE_FCVAR_INIT(VARNAME,0.0f)


/**
 * Create a floating-point shell variable.
 *
 * @note In order to use the shell variable, you still have to
 * tell GSH about it at runtime;
 * @see gsh_REGISTER()
 *
 * @param VARNAME the name of the variable
 * @param MIN     the minimum (and initial) value of the variable
 * @param MAX     the maximum value the variable can have
 * @param HELP    the help text associated with the variable
 */
#define gsh_DECLARE_FCVAR_BOUNDED_IMPL(VARNAME,MIN,MAX)                 \
    float   CVAR_##VARNAME = MIN;                                       \
    static void CMD_##VARNAME(int argc, char **argv)                    \
    {                                                                   \
        if (argc == 2)                                                  \
        {                                                               \
            float newval = (float)atof(argv[1]);                        \
            if (newval < MIN || newval > MAX) {                         \
                gsh_Printf("Error: " #VARNAME " must be "               \
                "in the range [%f,%f]\n",                               \
                           (float)MIN,(float)MAX);                      \
                return;                                                 \
            }                                                           \
            CVAR_##VARNAME = newval;                                    \
        }                                                               \
        else if (argc > 2)                                              \
        {                                                               \
            gsh_Printf(#VARNAME " [value]\n");                          \
            return;                                                     \
        }                                                               \
        gsh_Printf(#VARNAME " is %f\n", CVAR_##VARNAME);                \
    }


#define gsh_DECLARE_FCVAR_BOUNDED_EX(VARNAME,MIN,MAX,HELP)              \
    const char* HELP_##VARNAME = "(f) : " HELP " [" #MIN ".." #MAX "]"; \
    gsh_DECLARE_FCVAR_BOUNDED_IMPL(VARNAME,MIN,MAX)

#define gsh_DECLARE_FCVAR_BOUNDED(VARNAME,MIN,MAX)                      \
    const char* HELP_##VARNAME = "(f) : [" #MIN ".." #MAX "]";          \
    gsh_DECLARE_FCVAR_BOUNDED_IMPL(VARNAME,MIN,MAX)

/** use this to get an extern reference to a float CVAR */
#define gsh_ACCESS_FCVAR(VARNAME) \
    extern float CVAR_##VARNAME;
/** @} */

/** @name Integer CVARs */
/** @{ */
/**
 * Create an integer shell variable and initialize it.
 *
 * The user sets the value of shell variable "foo" interactively by 
 * typing "foo {n}", where {n} is the  value to which foo should be set.
 * To increment or decrement the value of foo, type "foo +" or "foo -", 
 * respectively.  
 * 
 * @note In order to use the shell variable, you still have to
 * tell GSH about it at runtime;
 * @see gsh_REGISTER()
 *
 * @param VARNAME the name of the variable
 * @param VALUE   the initial value of the variable
 * @param HELP    the help text associated with the variable
 */
#define gsh_DECLARE_ICVAR_INIT_IMPL(VARNAME,VALUE)          \
    int     CVAR_ ## VARNAME = VALUE;                       \
    static void CMD_ ## VARNAME (int argc, char **argv)     \
    {                                                       \
        if (argc == 2)                                      \
        {                                                   \
            if (strcmp(argv[1], "+") == 0)                  \
            {                                               \
                CVAR_ ## VARNAME += 1;                      \
            }                                               \
            else if (strcmp(argv[1], "-") == 0)             \
            {                                               \
                CVAR_ ## VARNAME -= 1;                      \
            }                                               \
            else                                            \
            {                                               \
                CVAR_ ## VARNAME = atoi(argv[1]);           \
            }                                               \
        }                                                   \
        else if (argc > 2)                                  \
        {                                                   \
            gsh_Printf(#VARNAME " [value]\n");              \
            return;                                         \
        }                                                   \
        gsh_Printf(#VARNAME " is %d\n", CVAR_ ## VARNAME);  \
    }


#define gsh_DECLARE_ICVAR_INIT_EX(VARNAME,VALUE,HELP)       \
    const char* HELP_##VARNAME = "(i) : " HELP;             \
    gsh_DECLARE_ICVAR_INIT_IMPL(VARNAME,VALUE)

#define gsh_DECLARE_ICVAR_INIT(VARNAME,VALUE)               \
    const char* HELP_##VARNAME = "(i)";                     \
    gsh_DECLARE_ICVAR_INIT_IMPL(VARNAME,VALUE)


/**
 * Create an integer shell variable and initialize it to zero.
 *
 * The user sets the value of shell variable "foo" interactively by 
 * typing "foo {n}", where {n} is the  value to which foo should be set.
 * To increment or decrement the value of foo, type "foo +" or "foo -", 
 * respectively.  
 * 
 * @note In order to use the shell variable, you still have to
 * tell GSH about it at runtime;
 * @see gsh_REGISTER()
 *
 * @param VARNAME the name of the variable
 */
#define gsh_DECLARE_ICVAR(VARNAME)                          \
    gsh_DECLARE_ICVAR_INIT(VARNAME,0)

#define gsh_DECLARE_ICVAR_EX(VARNAME,HELP)                  \
    gsh_DECLARE_ICVAR_INIT_EX(VARNAME,0,HELP)

/**
 * Create a bounded integer shell variable.
 *
 * The user sets the value of shell variable "foo" interactively by 
 * typing "foo {n}", where {n} is the  value to which foo should be set.
 * To increment or decrement the value of foo, type "foo +" or "foo -", 
 * respectively.  For bounded variables the value must be between MIN
 * and MAX.
 * 
 * Incrementing and decrementing "wrap around": for 
 * example, if incrementing foo would cause foo to exceed the maximum 
 * value MAX, foo is instead set to the minimum value MIN.
 *
 * @note In order to use the shell variable, you still have to
 * tell GSH about it at runtime;
 * @see gsh_REGISTER()
 *
 * @param VARNAME the name of the variable
 * @param MIN     the minimum (and initial) value of the variable
 * @param MAX     the maximum value the variable can have
 * @param HELP    the help text associated with the variable
 */
#define gsh_DECLARE_ICVAR_BOUNDED_IMPL(VARNAME,MIN,MAX)                 \
    int     CVAR_ ## VARNAME = MIN;                                     \
    static void CMD_ ## VARNAME (int argc, char **argv)                 \
    {                                                                   \
        int newval;                                                     \
        if (argc == 2)                                                  \
        {                                                               \
            if (strcmp(argv[1], "+") == 0)                              \
            {                                                           \
                newval = CVAR_ ## VARNAME + 1;                          \
                if (newval > MAX) { newval = MIN; } /* Wrap */          \
            }                                                           \
            else if (strcmp(argv[1], "-") == 0)                         \
            {                                                           \
                newval = CVAR_ ## VARNAME - 1;                          \
                if (newval < MIN) { newval = MAX; } /* Wrap */          \
            }                                                           \
            else                                                        \
            {                                                           \
                newval = atoi(argv[1]);                                 \
                if (newval < MIN || newval > MAX) {                     \
                    gsh_Printf("Error: " #VARNAME " must be "           \
                               "in the range [%d,%d]\n",MIN,MAX);       \
                    return;                                             \
                }                                                       \
            }                                                           \
            CVAR_ ## VARNAME = newval;                                  \
        }                                                               \
        else if (argc > 2)                                              \
        {                                                               \
            gsh_Printf(#VARNAME " [value]\n");                          \
            return;                                                     \
        }                                                               \
        gsh_Printf(#VARNAME " is %d\n", CVAR_ ## VARNAME);              \
    }


#define gsh_DECLARE_ICVAR_BOUNDED_EX(VARNAME,MIN,MAX,HELP)              \
    const char* HELP_##VARNAME = "(i) : " HELP " [" #MIN ".." #MAX "]"; \
    gsh_DECLARE_ICVAR_BOUNDED_IMPL(VARNAME,MIN,MAX)

#define gsh_DECLARE_ICVAR_BOUNDED(VARNAME,MIN,MAX)                      \
    const char* HELP_##VARNAME = "(i) : [" #MIN ".." #MAX "]";          \
    gsh_DECLARE_ICVAR_BOUNDED_IMPL(VARNAME,MIN,MAX)

/** use this to get an extern reference to an int CVAR */
#define gsh_ACCESS_ICVAR(VARNAME) \
    extern int CVAR_ ## VARNAME;
/** @} */

/** @name Enumerating CVARs */
/** @{ */
/**
 * Create a enumerating shell variable.
 *
 * The user sets the value of shell variable "foo" interactively by 
 * typing "foo {n}", where {n} is the value to which foo should be set,
 * in either numeric or symbolic form.  To step through the possible
 * values of foo, "foo +" and "foo -", just as with bounded ICVARs.
 *
 * Incrementing and decrementing "wrap around": for 
 * example, if incrementing foo would cause foo to pass the last 
 * value in the list, foo is instead set to the first value.
 *
 * The ECVAR defaults to the first entry in VALUES.
 *
 * @note In order to use the shell variable, you still have to
 * tell GSH about it at runtime;
 * @see gsh_REGISTER()
 *
 * @param VARNAME the name of the variable
 * @param VALUES  an array of gsh_ECVAR_VALUE_t's containing the valid values
 * @param HELP    the help text associated with the variable
 */
#define gsh_DECLARE_ECVAR_IMPL(VARNAME,VALUES)                                \
    int CVAR_##VARNAME = VALUES[0].val;                                       \
    int CVAR_##VARNAME##_IDX = 0;                                             \
    static void CMD_##VARNAME(int argc, char **argv)                          \
    {                                                                         \
        int newidx;                                                           \
                                                                              \
        if (argc == 2)                                                        \
        {                                                                     \
            int maxidx = _gsh_ECVAR_MAXIDX(VALUES);                           \
            if (strcmp(argv[1], "+") == 0)                                    \
            {                                                                 \
                newidx = CVAR_##VARNAME##_IDX + 1;                            \
                if (newidx > maxidx)                                          \
                {                                                             \
                    newidx = 0;                                               \
                }                                                             \
            }                                                                 \
            else if (strcmp(argv[1], "-") == 0)                               \
            {                                                                 \
                newidx = CVAR_##VARNAME##_IDX - 1;                            \
                if (newidx < 0)                                               \
                {                                                             \
                    newidx = maxidx;                                          \
                }                                                             \
            }                                                                 \
            else                                                              \
            {                                                                 \
                newidx = _gsh_ECVAR_NAMEFIND(VALUES, argv[1], maxidx);        \
                if (newidx == -1)                                             \
                {                                                             \
                    char *endptr;                                             \
                    errno = 0;                                                \
                    int e = (int)strtol(argv[1], &endptr, 10);                \
                    newidx = ((errno) || (endptr == argv[1]))                 \
                        ? -1 : _gsh_ECVAR_VALFIND(VALUES, e, maxidx);         \
                }                                                             \
                if (newidx < 0 || newidx > maxidx) {                          \
                    gsh_Printf("Error: Unknown value \"%s\" for "             \
                               #VARNAME "\n", argv[1]);                       \
                    return;                                                   \
                }                                                             \
            }                                                                 \
            CVAR_##VARNAME = VALUES[newidx].val;                              \
            CVAR_##VARNAME##_IDX = newidx;                                    \
        }                                                                     \
        else if (argc > 2)                                                    \
        {                                                                     \
            gsh_Printf(#VARNAME " [value]\n");                                \
            return;                                                           \
        }                                                                     \
        gsh_Printf(#VARNAME " is %s (%d)\n",                                  \
                   VALUES[CVAR_##VARNAME##_IDX].name, CVAR_##VARNAME);        \
    }

#define gsh_DECLARE_ECVAR_EX(VARNAME,VALUES,HELP)                             \
    const char* HELP_##VARNAME = "(e) : " HELP;                               \
    gsh_DECLARE_ECVAR_IMPL(VARNAME,VALUES)

#define gsh_DECLARE_ECVAR(VARNAME,VALUES)                                     \
    const char* HELP_##VARNAME = "(e)";                                       \
    gsh_DECLARE_ECVAR_IMPL(VARNAME,VALUES)

/**
 * declare an array of these for each of your enumerated CVARs.
 * the first value in the list will be the default value for the CVAR.
 *
 * eg:
 *
 * static const gsh_ECVAR_VALUE_t foovals[] = {
 *                                              {0, "foo"},
 *                                              {1, "bar"},
 *                                              {5, "baz"}
 *                                            };
 */
typedef struct { int val; const char *name; } gsh_ECVAR_VALUE_t;

int _gsh_ECVAR_NAMEFIND(const gsh_ECVAR_VALUE_t *values,
                        const char *name, int maxidx);

int _gsh_ECVAR_VALFIND(const gsh_ECVAR_VALUE_t *values,
                       int e, int maxidx);

#define _gsh_ECVAR_MAXIDX(VALUES) \
    (sizeof(VALUES) / sizeof(gsh_ECVAR_VALUE_t) - 1)


/** use this to get an extern reference to an enum CVAR */
#define gsh_ACCESS_ECVAR(VARNAME) \
    extern int CVAR_ ## VARNAME;
/** @} */

/** @name Boolean CVARs */
/** @{ */
/**
 * Create a boolean shell variable.
 *
 * The user sets the value of shell variable "foo" interactively by 
 * typing "foo {bool}", where {bool} is the value to which foo should 
 * be set, and is either 0 or 1.  In the shell the user may also type
 * "false" (equivalent to 0), or "true" (equivalent to 1).
 * 
 * To toggle the value of foo between 0 and 1, type "foo +".
 * 
 * @note In order to use the shell variable, you still have to
 * tell GSH about it at runtime;
 * @see gsh_REGISTER()
 *
 * @param VARNAME the name of the variable
 * @param VALUE the initial value of the variable (0 or 1)
 */

#define gsh_DECLARE_BCVAR_EX(VARNAME,VALUE,HELP)                              \
    const char* HELP_##VARNAME = "(b) : " HELP;                               \
    gsh_DECLARE_ECVAR_IMPL(VARNAME,_gsh_boolvals##VALUE);

#define gsh_DECLARE_BCVAR(VARNAME,VALUE)                                      \
    const char* HELP_##VARNAME = "(b)";                                       \
    gsh_DECLARE_ECVAR_IMPL(VARNAME,_gsh_boolvals##VALUE);

static const gsh_ECVAR_VALUE_t _gsh_boolvals0[] = { {0,"false"},{1,"true"} };
static const gsh_ECVAR_VALUE_t _gsh_boolvals1[] = { {1,"true"},{0,"false"} };

/** use this to get an extern reference to a boolean CVAR */
#define gsh_ACCESS_BCVAR(VARNAME) \
    extern int CVAR_ ## VARNAME;
/** @} */

/** @name String CVARs */
/** @{ */
#define gsh_DECLARE_SCVAR_INIT_IMPL(VARNAME,VALUE)                      \
    char*   CVAR_ ## VARNAME = VALUE;                                   \
    int     CVAR_ ## VARNAME ## _FREEABLE = 0;                          \
    static void CMD_ ## VARNAME (int argc, char **argv)                 \
    {                                                                   \
        if (argc == 2)                                                  \
        {                                                               \
            if (CVAR_ ## VARNAME && CVAR_ ## VARNAME ## _FREEABLE) {    \
                free(CVAR_ ## VARNAME);                                 \
            }                                                           \
            CVAR_ ## VARNAME = strdup(argv[1]);                         \
            CVAR_ ## VARNAME ## _FREEABLE = 1;                          \
        }                                                               \
        else if (argc > 2)                                              \
        {                                                               \
            gsh_Printf(#VARNAME " [value]\n");                          \
            return;                                                     \
        }                                                               \
        gsh_Printf(#VARNAME " is \"%s\"\n", CVAR_ ## VARNAME);          \
    }

#define gsh_DECLARE_SCVAR_INIT_EX(VARNAME,VALUE,HELP)                   \
    const char* HELP_ ## VARNAME = "(s) : " HELP;                       \
    gsh_DECLARE_SCVAR_INIT_IMPL(VARNAME,VALUE)

#define gsh_DECLARE_SCVAR_INIT(VARNAME,VALUE)                           \
    const char* HELP_ ## VARNAME = "(s)";                               \
    gsh_DECLARE_SCVAR_INIT_IMPL(VARNAME,VALUE)

#define gsh_DECLARE_SCVAR(VARNAME)                                      \
    gsh_DECLARE_SCVAR_INIT(VARNAME,NULL)

#define gsh_DECLARE_SCVAR_EX(VARNAME,HELP)                              \
    gsh_DECLARE_SCVAR_INIT_EX(VARNAME,NULL,HELP)

#define gsh_ACCESS_SCVAR(VARNAME) \
    extern char* CVAR_ ## VARNAME;
/** @} */

/** @name Accessors and helpers */
/** @{ */
#define gsh_CVAR(VARNAME) \
    CVAR_ ## VARNAME

#define gsh_CMD(VARNAME) \
    CMD_ ## VARNAME

#define gsh_DECLARE_CMD_IMPL(CMDNAME)                                   \
	static void CMD_ ## CMDNAME (int argc, char **argv)


#define gsh_DECLARE_CMD_EX(CMDNAME,HELP)                                \
    const char* HELP_ ## CMDNAME = "cmd : " HELP;                       \
	gsh_DECLARE_CMD_IMPL(CMDNAME)

#define gsh_DECLARE_CMD(CMDNAME)                                        \
    const char* HELP_ ## CMDNAME = "cmd";                               \
	gsh_DECLARE_CMD_IMPL(CMDNAME)

#define gsh_DECLARE_SYNONYM_IMPL(CMDNAME,SYNONYM)                       \
    static void CMD_ ## SYNONYM (int argc, char **argv)                 \
    {                                                                   \
        CMD_ ## CMDNAME(argc, argv);                                    \
    }

#define gsh_DECLARE_SYNONYM(CMDNAME,SYNONYM)                            \
    const char* HELP_ ## SYNONYM = "syn : alias for '" #CMDNAME "'";    \
    gsh_DECLARE_SYNONYM_IMPL(CMDNAME,SYNONYM)
    
#define gsh_REGISTER(NAME) \
    {gsh_AddCommand(#NAME, CMD_ ## NAME, HELP_ ## NAME);}

#define gsh_REGISTER_HIDDEN(NAME) \
    {gsh_AddHiddenCommand(#NAME, CMD_ ## NAME, HELP_ ## NAME);}

/** @} */

/** @name Core API */
/** @{ */
void   gsh_Init(void (*pfnProc)(const char *));
void   gsh_RC(void);
void   gsh_Shutdown(void);
void   gsh_WriteToFile(FILE *pfdOutput);

int    gsh_Printf(const char *szFormat, ...);
void   gsh_AddCommand(const char *szName, void (*pfnProc)(int, char **), const char* szHelp);
void   gsh_AddHiddenCommand(const char *szName, void (*pfnProc)(int, char **), const char* szHelp);
int    gsh_AddAlias(const char *szCmdName, const char *szAliasName);
void   gsh_ExecLine(const char *szLine);
int    gsh_RunCommand(int argc, char **argv);
const char *gsh_ListCommands(const char *szKey);

void   gsh_SetValStr(char *szName, char *szValue);
/** @} */

#ifdef __cplusplus
}
#endif

#endif

/** @} */


